"""Generate C++ code of common functions by including the code directly."""

# pylint: disable=line-too-long

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    common as cpp_common,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_concatenate_definitions_for_2_parts_and_above() -> List[Stripped]:
    """
    Generate the definition of ``concat`` functions.

    >>> print(_generate_concatenate_definitions_for_2_parts_and_above()[0])
    /**
     * Concatenate 2 strings.
     *
     * \\param part0 1st part of the concatenation
     * \\param part1 2nd part of the concatenation
     * \\return a concatenation of the 2 parts
     */
    std::string Concat(
      const std::string& part0,
      const std::string& part1
    );
    """
    concat_funcs = []  # type: List[Stripped]
    for string_type in ["std::string", "std::wstring"]:
        for count in range(2, 65):
            param_descriptions = []  # type: List[str]
            for i in range(0, count):
                ordinal: str

                ordinal_i = i + 1

                if ordinal_i % 10 == 1:
                    ordinal = f"{ordinal_i}st"
                elif ordinal_i % 10 == 2:
                    ordinal = f"{ordinal_i}nd"
                elif ordinal_i % 10 == 3:
                    ordinal = f"{ordinal_i}rd"
                else:
                    ordinal = f"{ordinal_i}th"

                param_descriptions.append(
                    f" * \\param part{i} {ordinal} part of the concatenation"
                )
            param_description_joined = "\n".join(param_descriptions)

            args_definition = ",\n".join(
                f"{I}const {string_type}& part{i}" for i in range(0, count)
            )

            concat_funcs.append(
                Stripped(
                    f"""\
/**
 * Concatenate {count} strings.
 *
{param_description_joined}
 * \\return a concatenation of the {count} parts
 */
{string_type} Concat(
{args_definition}
);"""
                )
            )

    return concat_funcs


def _generate_concatenate_implementations_for_2_parts_and_above() -> List[Stripped]:
    """
    Generate the implementation of ``concat`` functions.

    >>> print(_generate_concatenate_implementations_for_2_parts_and_above()[0])
    std::string Concat(
      const std::string& part0,
      const std::string& part1
    ) {
      size_t size = 0;
      size += part0.size();
      size += part1.size();
    <BLANKLINE>
      std::string result;
      result.reserve(size);
    <BLANKLINE>
      result.append(part0);
      result.append(part1);
    <BLANKLINE>
      return result;
    }
    """
    concat_funcs = []  # type: List[Stripped]
    for string_type in ["std::string", "std::wstring"]:
        for count in range(2, 65):
            args_definition = ",\n".join(
                f"{I}const {string_type}& part{i}" for i in range(0, count)
            )

            size_block = Stripped(
                "\n".join(f"size += part{i}.size();" for i in range(0, count))
            )

            append_block = Stripped(
                "\n".join(f"result.append(part{i});" for i in range(0, count))
            )

            concat_funcs.append(
                Stripped(
                    f"""\
{string_type} Concat(
{args_definition}
) {{
{I}size_t size = 0;
{I}{indent_but_first_line(size_block, I)}

{I}{string_type} result;
{I}result.reserve(size);

{I}{indent_but_first_line(append_block, I)}

{I}return result;
}}"""
                )
            )

    return concat_funcs


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(library_namespace: Stripped) -> str:
    """Generate the C++ header code for common functions."""
    namespace = Stripped(f"{library_namespace}")

    include_guard_var = cpp_common.include_guard_var(namespace)

    make_uniques = [
        Stripped(
            f"""\
template<typename T>
std::unique_ptr<T> make_unique() {{
{I}return std::unique_ptr<T>(
{II}new T()
{I});
}}"""
        )
    ]  # type: List[Stripped]
    for i in range(1, 16):
        typenames_joined = ",\n".join(
            ["typename T"] + [f"typename A{j + 1}" for j in range(i)]
        )

        args_joined = ",\n".join(f"A{j + 1}&& a{j + 1}" for j in range(i))

        forward_args_joined = ",\n".join(
            f"std::forward<A{j + 1}>(a{j + 1})" for j in range(i)
        )

        make_uniques.append(
            Stripped(
                f"""\
template<
{I}{indent_but_first_line(typenames_joined, I)}
>
std::unique_ptr<T> make_unique(
{I}{indent_but_first_line(args_joined, I)}
) {{
{I}return std::unique_ptr<T>(
{II}new T(
{III}{indent_but_first_line(forward_args_joined, III)}
{II})
{I});
}}"""
            )
        )

    make_uniques_joined = "\n\n".join(make_uniques)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            """\
#pragma warning(push, 0)
#include <algorithm>
#include <functional>
#include <memory>
#include <sstream>
#include <string>
#include <utility>
#pragma warning(pop)

// See: https://stackoverflow.com/questions/2324658/how-to-determine-the-version-of-the-c-standard-used-by-the-compiler
#if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
// Standard library provides std::optional in C++17 and above.
#pragma warning(push, 0)
#include <optional>
#pragma warning(pop)
#else
// We rely on https://github.com/TartanLlama/optional for optional structure.
#pragma warning(push, 0)
#include <tl/optional.hpp>
#pragma warning(pop)
#endif

// NOTE (mristin):
// We check for the version above C++20 as there is no C++23 literal yet, and
// std::expected is available only in C++23.
// See: https://stackoverflow.com/questions/2324658/how-to-determine-the-version-of-the-c-standard-used-by-the-compiler
// and: http://eel.is/c++draft/cpp.predefined#1.1
#if ((defined(_MSVC_LANG) && _MSVC_LANG > 202002L) || __cplusplus > 202002L)
// Standard library provides std::expected in C++23 and above.
#pragma warning(push, 0)
#include <expected>
#pragma warning(pop)
#else
// We rely on https://github.com/TartanLlama/expected for expected structure.
#pragma warning(push, 0)
#include <tl/expected.hpp>
#pragma warning(pop)
#endif"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            f"""\
/**
 * \\defgroup common Common functionality used throughout the library
 * @{{
 */
namespace {cpp_common.COMMON_NAMESPACE} {{"""
        ),
        Stripped(
            """\
// Please keep in sync with the preprocessing directives above in the include block.
#if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
// Standard library provides std::optional in C++17 and above.
using std::optional;
using std::nullopt;
using std::make_optional;
#else
using tl::optional;
using tl::nullopt;
using tl::make_optional;
#endif"""
        ),
        Stripped(
            """\
// Please keep in sync with the preprocessing directives above in the include block.
#if ((defined(_MSVC_LANG) && _MSVC_LANG > 202002L) || __cplusplus > 202002L)
using std::expected;
using std::unexpected;
using std::make_unexpected;
#else
using tl::expected;
using tl::unexpected;
using tl::make_unexpected;
#endif"""
        ),
        Stripped(
            f"""\
// Please keep in sync with the preprocessing directives above in the include block.
// Standard library provides std::make_unique in C++14 and above.
#if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201402L) || __cplusplus >= 201402L)
using std::make_unique;
#else
// Inspired by:
// https://stackoverflow.com/questions/12547983/is-there-a-way-to-write-make-unique-in-vs2012
{make_uniques_joined}
#endif"""
        ),
        *_generate_concatenate_definitions_for_2_parts_and_above(),
        Stripped(
            f"""\
/**
 * Check if all the elements satisfy the \\p condition.
 *
 * \\param condition returning a boolean to be checked for each element
 * \\param container to be iterated through
 * \\return `true` if all the elements of \\p container satisfy the \\p condition
 */
template<typename ContainerT, typename FunctorT>
bool All(
{I}FunctorT condition,
{I}const ContainerT& container
) {{
{I}for (const auto& item : container ) {{
{II}if (!condition(item)) {{
{III}return false;
{II}}}
{I}}}
{I}return true;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check if any of the elements satisfy the \\p condition.
 *
 * \\param condition returning a boolean to be checked for each element
 * \\param container to be iterated through
 * \\return `true` if any of the elements of \\p container satisfy the \\p condition
 */
template<typename ContainerT, typename FunctorT>
bool Some(
{I}FunctorT condition,
{I}const ContainerT& container
) {{
{I}for (const auto& item : container ) {{
{II}if (condition(item)) {{
{III}return true;
{II}}}
{I}}}
{I}return false;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check if all the numbers in the range `[start, end)` satisfy the \\p condition.
 *
 * \\param condition returning a boolean to be checked for each number
 * \\param start of the range
 * \\param end of the range
 * \\return \\parblock
 * `true` if all the numbers between \\p start and \\p end (excluded)
 * satisfy the \\p condition
 * \\endparblock
 */
template<typename FunctorT>
bool AllRange(
{I}FunctorT condition,
{I}size_t start,
{I}size_t end
) {{
{I}for (size_t i = start; i < end; ++i) {{
{II}if (!condition(i)) {{
{III}return false;
{II}}}
{I}}}
{I}return true;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check if any number in the range `[start, end)` satisfy the \\p condition.
 *
 * \\param condition returning a boolean to be checked for each number
 * \\param start of the range
 * \\param end of the range
 * \\return \\parblock
 * `true` if any number between \\p start and
 * \\p end (excluded) satisfy the \\p condition
 * \\endparblock
 */
template<typename FunctorT>
bool SomeRange(
{I}FunctorT condition,
{I}size_t start,
{I}size_t end
) {{
{I}for (size_t i = start; i < end; ++i) {{
{II}if (condition(i)) {{
{III}return true;
{II}}}
{I}}}
{I}return false;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check if the \\p container contains the \\p value.
 *
 * \\param container to be searched through
 * \\param value to be search for
 * \\return `true` if \\p value is in the \\p container
 */
template<typename ContainerT, typename ValueT>
bool Contains(
{I}const ContainerT& container,
{I}const ValueT& value
) {{
{I}const auto container_begin = std::begin(container);
{I}const auto container_end = std::end(container);
{I}return std::find(
{II}container_begin,
{II}container_end,
{II}value
{I}) != container_end;
}}"""
        ),
        Stripped(
            """\
/**
 * Convert platform-independent the wide string to a UTF-8 string.
 *
 * \\param text to be converted
 * \\return UTF-8 encoded \\p text
 */
std::string WstringToUtf8(const std::wstring& text);"""
        ),
        Stripped(
            f"""\
/**
 * Convert platform-independent the UTF-8 encoded string to a wide string.
 *
 * \\param utf8_text UTF-8 encoded text to be converted
 * \\param utf8_text_size size of the text to be converted. If std::string::npos,
 * the \\p utf8_text is assumed to be null-terminated and the size is determined
 * using `strlen`.
 * \\return the wide-string representation
 */
std::wstring Utf8ToWstring(
{I}const char* utf8_text,
{I}size_t utf8_text_size = std::string::npos
);"""
        ),
        Stripped(
            """\
/**
 * Convert platform-independent the UTF-8 encoded string to a wide string.
 *
 * \\param utf8_text UTF-8 encoded text to be converted
 * \\return wide string
 */
std::wstring Utf8ToWstring(const std::string& utf8_text);"""
        ),
        Stripped(
            f"""\
}}  // namespace {cpp_common.COMMON_NAMESPACE}
/**@}}*/"""
        ),
        cpp_common.generate_namespace_closing(library_namespace),
        cpp_common.WARNING,
        Stripped(f"#endif  // {include_guard_var}"),
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(library_namespace: Stripped) -> str:
    """Generate the C++ code for common functions."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.COMMON_NAMESPACE}")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f'''\
#include "{include_prefix_path}/common.hpp"'''
        ),
        Stripped(
            """\
#pragma warning(push, 0)
#include <algorithm>
#pragma warning(pop)

// NOTE (mristin):
// We use MultiByteToWideChar and WideCharToMulitByte from <windows.h> on Windows
// as std::codecvt is not robust enough on Windows, and produces invalid UTF-16
// sequences as std::wstring. See:
// https://stackoverflow.com/questions/2573834/c-convert-string-or-char-to-wstring-or-wchar-t,
// especially the comment:
// https://stackoverflow.com/questions/2573834/c-convert-string-or-char-to-wstring-or-wchar-t#comment110447503_18597384
#ifdef _WIN32
#pragma warning(push, 0)
#include <windows.h>
#include <limits>
#include <cstring>
#pragma warning(pop)
#else
// NOTE (mristin):
// We use codecvt although it has been deprecated. There has been not
// suitable replacement proposed yet, so we simply stick to it
// until there is. See:
// https://stackoverflow.com/questions/42946335/deprecated-header-codecvt-replacement

#pragma warning(push, 0)
#include <codecvt>
#include <locale>
#pragma warning(pop)
#endif"""
        ),
        cpp_common.generate_namespace_opening(namespace),
        *_generate_concatenate_implementations_for_2_parts_and_above(),
        Stripped(
            f"""\
std::string WstringToUtf8(const std::wstring& text) {{
{I}#ifdef _WIN32
{I}// Inspired by:
{I}// https://stackoverflow.com/a/69410299/1600678

{I}if (text.empty()) {{
{II}return "";
{I}}}

{I}// NOTE (mristin):
{I}// We put `max` into parentheses to avoid conflicts with
{I}// <windows.h> `max` macro, see:
{I}// https://stackoverflow.com/questions/11544073/how-do-i-deal-with-the-max-macro-in-windows-h-colliding-with-max-in-std

{I}const size_t text_size = text.size();
{I}if (
{II}text_size
{II}> static_cast<size_t>(
{III}(std::numeric_limits<int>::max)()
{II})
{I}) {{
{II}throw std::out_of_range(
{III}common::Concat(
{IIII}"The size of the text to be converted to UTF-8, ",
{IIII}std::to_string(text_size),
{IIII}", exceeds the maximum int value ",
{IIII}std::to_string((std::numeric_limits<int>::max)())
{III})
{II});
{I}}}
{I}const int text_size_int = static_cast<int>(text_size);

{I}const auto size_needed = WideCharToMultiByte(
{II}CP_UTF8,
{II}0,
{II}&(text[0]),
{II}text_size_int,
{II}nullptr,
{II}0,
{II}nullptr,
{II}nullptr
{I});

{I}if (size_needed <= 0) {{
{II}throw std::runtime_error(
{III}"WideCharToMultiByte() failed: " + std::to_string(size_needed)
{II});
{I}}}

{I}std::string result(size_needed, 0);

{I}WideCharToMultiByte(
{II}CP_UTF8,
{II}0,
{II}&(text[0]),
{II}text_size_int,
{II}&(result[0]),
{II}size_needed,
{II}nullptr,
{II}nullptr
{I});

{I}return result;
{I}#else
{I}// NOTE (mristin):
{I}// We use codecvt although it has been deprecated. There has been not
{I}// suitable replacement proposed yet, so we simply stick to it
{I}// until there is. See:
{I}// https://stackoverflow.com/questions/42946335/deprecated-header-codecvt-replacement

{I}std::wstring_convert<std::codecvt_utf8<wchar_t> > conv;
{I}return conv.to_bytes(text.data());
{I}#endif
}}"""
        ),
        Stripped(
            f"""\
std::wstring Utf8ToWstring(
{I}const char* utf8_text,
{I}size_t utf8_text_size
) {{
{I}if (utf8_text_size == 0) {{
{II}return std::wstring();
{I}}}

{I}#ifdef _WIN32
{I}// NOTE (mristin):
{I}// We have to use MultiByteToWideChar from <windows.h> on Windows
{I}// as std::codecvt is not robust enough on Windows and produces invalid UTF-16
{I}// sequences as std::wstring. See:
{I}// https://stackoverflow.com/questions/2573834/c-convert-string-or-char-to-wstring-or-wchar-t,
{I}// especially the comment:
{I}// https://stackoverflow.com/questions/2573834/c-convert-string-or-char-to-wstring-or-wchar-t#comment110447503_18597384

{I}// Inspired by:
{I}// https://stackoverflow.com/a/69410299/1600678

{I}if (utf8_text_size == std::string::npos) {{
{II}utf8_text_size = strlen(utf8_text);
{I}}}

{I}// NOTE (mristin):
{I}// We put `max` into parentheses to avoid conflicts with
{I}// <windows.h> `max` macro, see:
{I}// https://stackoverflow.com/questions/11544073/how-do-i-deal-with-the-max-macro-in-windows-h-colliding-with-max-in-std

{I}if (
{II}utf8_text_size
{II}> static_cast<size_t>(
{III}(std::numeric_limits<int>::max)()
{II})
{I}) {{
{II}throw std::out_of_range(
{III}common::Concat(
{IIII}"The size of the UTF-8 text to be converted to wide string, ",
{IIII}std::to_string(utf8_text_size),
{IIII}", exceeds the maximum int value ",
{IIII}std::to_string((std::numeric_limits<int>::max)())
{III})
{II});
{I}}}
{I}const int utf8_text_size_int = static_cast<int>(utf8_text_size);

{I}const auto size_needed = MultiByteToWideChar(
{II}CP_UTF8,
{II}0,
{II}utf8_text,
{II}utf8_text_size_int,
{II}nullptr,
{II}0
{I});
{I}if (size_needed <= 0) {{
{II}throw std::runtime_error(
{III}"MultiByteToWideChar() failed: " + std::to_string(size_needed)
{II});
{I}}}

{I}std::wstring result(size_needed, 0);

{I}MultiByteToWideChar(
{II}CP_UTF8,
{II}0,
{II}utf8_text,
{II}utf8_text_size_int,
{II}&(result[0]),
{II}size_needed
{I});

{I}return result;

{I}#else
{I}// NOTE (mristin):
{I}// We use codecvt although it has been deprecated. There has been not
{I}// suitable replacement proposed yet, so we simply stick to it
{I}// until there is. See:
{I}// https://stackoverflow.com/questions/42946335/deprecated-header-codecvt-replacement

{I}std::wstring_convert<std::codecvt_utf8<wchar_t>,wchar_t> conv;
{I}return conv.from_bytes(utf8_text);
{I}#endif
}}"""
        ),
        Stripped(
            f"""\
std::wstring Utf8ToWstring(const std::string& utf8_text) {{
{I}return Utf8ToWstring(&(utf8_text[0]), utf8_text.size());
}}"""
        ),
        cpp_common.generate_namespace_closing(namespace),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Render descriptions to C# documentation comments."""
import abc
import collections
import io
import itertools
import textwrap
import xml.sax.saxutils
from typing import (
    Tuple,
    Optional,
    List,
    OrderedDict,
    Union,
    Sequence,
    TypeVar,
    Iterator,
    Iterable,
)

import docutils.nodes
import docutils.parsers.rst.roles
import docutils.utils
from icontract import require, ensure, DBC

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Error,
    assert_never,
    Identifier,
    assert_union_of_descendants_exhaustive,
    assert_union_without_excluded,
)
from aas_core_codegen.csharp import (
    naming as csharp_naming,
)
from aas_core_codegen.intermediate import (
    doc as intermediate_doc,
    _translate as intermediate_translate,
)


class _Node(DBC):
    """Represent a node in an AST of a documentation comment."""

    @abc.abstractmethod
    def accept(self, visitor: "_NodeVisitor") -> None:
        """Accept the ``visitor`` and dispatch."""
        raise NotImplementedError()


class _Text(_Node):
    """Represent a text node in a documentation comment."""

    def __init__(self, content: str) -> None:
        """Initialize with the given values."""
        self.content = content

    def accept(self, visitor: "_NodeVisitor") -> None:
        """Accept the ``visitor`` and dispatch."""
        visitor.visit_text(self)

    def __repr__(self) -> str:
        """Generate a string representation for easier debugging."""
        return f"{self.__class__.__name__}({self.content!r})"


class _List(_Node):
    """
    Represent a sequence of nodes of a C# documentation comment.

    This is necessary so that we can render a concatenation where there is no
    enclosing element.
    """

    def __init__(self, items: List["_NodeUnion"]) -> None:
        self.items = items

    def accept(self, visitor: "_NodeVisitor") -> None:
        """Accept the ``visitor`` and dispatch."""
        visitor.visit_list(self)

    def __repr__(self) -> str:
        """Generate a string representation for easier debugging."""
        if len(self.items) == 0:
            return f"{self.__class__.__name__}([])"

        writer = io.StringIO()
        writer.write(f"{self.__class__.__name__}(\n")
        writer.write("  [\n")

        for i, item in enumerate(self.items):
            if i > 0:
                writer.write(",\n")
            writer.write(textwrap.indent(repr(item), "    "))

        writer.write("\n  ]\n)")
        return writer.getvalue()


class _Element(_Node):
    """Represent an element of a C# documentation comment."""

    def __init__(
        self,
        name: str,
        attrs: Optional[OrderedDict[str, str]] = None,
        children: Optional[_List] = None,
    ) -> None:
        self.name = name
        self.attrs = collections.OrderedDict() if attrs is None else attrs
        self.children = _List(items=[]) if children is None else children

    def accept(self, visitor: "_NodeVisitor") -> None:
        """Accept the ``visitor`` and dispatch."""
        visitor.visit_element(self)

    def __repr__(self) -> str:
        """Generate a string representation for easier debugging."""
        indented_name = textwrap.indent(repr(self.name), "  ")
        indented_attrs = textwrap.indent(repr(self.attrs), "  ")
        indented_children = textwrap.indent(repr(self.children), "  ")

        return f"""\
{self.__class__.__name__}(
{indented_name},
{indented_attrs},
{indented_children}
)"""


_NodeUnion = Union[_Text, _Element, _List]
assert_union_of_descendants_exhaustive(base_class=_Node, union=_NodeUnion)


class _NodeVisitor:
    def visit(self, node: _Node) -> None:
        """Visit *via* double-dispatch."""
        node.accept(self)

    def visit_text(self, node: _Text) -> None:
        """Visit the text node."""
        pass

    def visit_list(self, node: _List) -> None:
        """Visit the node list and its items recursively."""
        for item in node.items:
            self.visit(item)

    def visit_element(self, node: _Element) -> None:
        """Visit the element node and its children."""
        self.visit(node.children)


class _ElementRenderer(intermediate_doc.DocutilsElementTransformer[_NodeUnion]):
    """Render descriptions as C# docstring XML."""

    def transform_text(
        self, element: docutils.nodes.Text
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        return _Text(element.astext()), None

    def transform_reference_to_our_type_in_doc(
        self, element: intermediate_doc.ReferenceToOurType
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        name: str

        if isinstance(element.our_type, intermediate.Enumeration):
            name = csharp_naming.enum_name(element.our_type.name)

        elif isinstance(element.our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2021-12-17):
            # We do not generate a class for constrained primitives, but we
            # leave it as class name, as that is what we used for ``Verify*`` function.
            name = csharp_naming.class_name(element.our_type.name)

        elif isinstance(element.our_type, intermediate.Class):
            if isinstance(element.our_type, intermediate.AbstractClass):
                # NOTE (mristin, 2021-12-25):
                # We do not generate C# code for abstract classes, so we have to refer
                # to the interface.
                name = csharp_naming.interface_name(element.our_type.name)

            elif isinstance(element.our_type, intermediate.ConcreteClass):
                # NOTE (mristin, 2021-12-25):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.
                name = csharp_naming.class_name(element.our_type.name)

            else:
                assert_never(element.our_type)

        else:
            # NOTE (mristin, 2022-03-30):
            # This is a very special case where we had problems with an interface.
            # We leave this check here, just in case the bug resurfaces.
            if isinstance(element.our_type, intermediate_translate._PlaceholderOurType):
                return None, [
                    f"Unexpected placeholder for our type: {element.our_type}; "
                    f"this is a bug"
                ]

            assert_never(element.our_type)

        # NOTE (mristin, 2022-06-19):
        # We need to prefix the cref in case there are naming conflicts.
        prefixed_name = f"Aas.{name}"

        return (
            _Element(
                name="see", attrs=collections.OrderedDict([("cref", prefixed_name)])
            ),
            None,
        )

    def transform_reference_to_attribute_in_doc(
        self, element: intermediate_doc.ReferenceToAttribute
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        cref: str

        if isinstance(element.reference, intermediate_doc.ReferenceToProperty):
            name_of_our_type: str

            if isinstance(element.reference.cls, intermediate.AbstractClass):
                # We do not generate C# code for abstract classes, so we have to refer
                # to the interface.
                name_of_our_type = csharp_naming.interface_name(
                    element.reference.cls.name
                )
            elif isinstance(element.reference.cls, intermediate.ConcreteClass):
                # NOTE (mristin, 2021-12-25):
                # Though a concrete class can have multiple descendants and the writer
                # might actually want to refer to the *interface* instead of
                # the concrete class, we do the best effort here and resolve it to the
                # name of the concrete class.

                name_of_our_type = csharp_naming.class_name(element.reference.cls.name)
            else:
                assert_never(element.reference.cls)

            prop_name = csharp_naming.property_name(element.reference.prop.name)

            cref = f"{name_of_our_type}.{prop_name}"
        elif isinstance(
            element.reference, intermediate_doc.ReferenceToEnumerationLiteral
        ):
            name_of_our_type = csharp_naming.enum_name(
                element.reference.enumeration.name
            )
            literal_name = csharp_naming.enum_literal_name(
                element.reference.literal.name
            )

            cref = f"{name_of_our_type}.{literal_name}"
        else:
            # NOTE (mristin, 2022-03-30):
            # This is a very special case where we had problems with an interface.
            # We leave this check here, just in case the bug resurfaces.
            if isinstance(
                element.reference,
                intermediate_translate._PlaceholderReferenceToAttribute,
            ):
                return None, [
                    f"Unexpected placeholder "
                    f"for the attribute reference: {element.reference}; "
                    f"this is a bug"
                ]

            assert_never(element.reference)

        # NOTE (mristin, 2022-06-19):
        # We need to prefix the cref in case there are naming conflicts.
        prefixed_cref = f"Aas.{cref}"

        return (
            _Element(
                name="see", attrs=collections.OrderedDict([("cref", prefixed_cref)])
            ),
            None,
        )

    def transform_reference_to_argument_in_doc(
        self, element: intermediate_doc.ReferenceToArgument
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        arg_name = csharp_naming.argument_name(Identifier(element.reference))

        return (
            _Element(
                name="paramref", attrs=collections.OrderedDict([("name", arg_name)])
            ),
            None,
        )

    def transform_reference_to_constraint_in_doc(
        self, element: intermediate_doc.ReferenceToConstraint
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        return _Text(content=f"Constraint {element.reference}"), None

    def transform_reference_to_constant_in_doc(
        self, element: intermediate_doc.ReferenceToConstant
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        constant_as_prop_name = csharp_naming.property_name(element.constant.name)
        cref = f"Aas.Constants.{constant_as_prop_name}"

        return (
            _Element(name="see", attrs=collections.OrderedDict([("cref", cref)])),
            None,
        )

    def transform_literal(
        self, element: docutils.nodes.literal
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        return (
            _Element(name="c", children=_List(items=[_Text(content=element.astext())])),
            None,
        )

    def _transform_children_of(
        self,
        element: docutils.nodes.Element,
    ) -> Tuple[Optional[_List], Optional[List[str]]]:
        """Transform the children to a Python list."""
        children = []  # type: List[_NodeUnion]

        errors = []  # type: List[str]
        for child in element.children:
            rendered_child, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert rendered_child is not None
                children.append(rendered_child)

        if len(errors) > 0:
            return None, errors

        return _List(items=children), None

    def transform_paragraph(
        self, element: docutils.nodes.paragraph
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        children, errors = self._transform_children_of(element)
        if errors is not None:
            return None, errors

        assert children is not None

        return _Element(name="para", children=children), None

    def transform_emphasis(
        self, element: docutils.nodes.emphasis
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        children, errors = self._transform_children_of(element)
        if errors is not None:
            return None, errors

        assert children is not None

        return _Element(name="em", children=children), None

    def transform_list_item(
        self, element: docutils.nodes.list_item
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        children, errors = self._transform_children_of(element)
        if errors is not None:
            return None, errors

        assert children is not None

        return _Element(name="li", children=children), None

    def transform_bullet_list(
        self, element: docutils.nodes.bullet_list
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        children, errors = self._transform_children_of(element)
        if errors is not None:
            return None, errors

        assert children is not None

        return _Element(name="ul", children=children), None

    def transform_note(
        self, element: docutils.nodes.note
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        children, errors = self._transform_children_of(element)

        if errors is not None:
            return None, errors

        assert children is not None

        return _Element(name="para", children=children), None

    def transform_reference(
        self, element: docutils.nodes.reference
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        return self._transform_children_of(element)

    def transform_field_body(
        self, element: docutils.nodes.field_body
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        return self._transform_children_of(element)

    def transform_document(
        self, element: docutils.nodes.document
    ) -> Tuple[Optional[_NodeUnion], Optional[List[str]]]:
        return self._transform_children_of(element)


class _FlattenListVisitor(_NodeVisitor):
    """Flatten all the node lists recursively and in-place."""

    def visit_list(self, node: _List) -> None:
        """Visit the node list and its items recursively."""
        new_items = []  # type: List[_NodeUnion]

        for item in node.items:
            self.visit(item)

            if isinstance(item, _List):
                new_items.extend(item.items)
            else:
                new_items.append(item)

        node.items = new_items


class _ConcatenateTextVisitor(_NodeVisitor):
    """Concatenate the consecutive text elements in lists recursively and in-place."""

    def visit_list(self, node: _List) -> None:
        """Visit the node list and its items recursively."""
        new_items = []  # type: List[_NodeUnion]

        accumulator = []  # type: List[_Text]
        for item in node.items:
            if isinstance(item, _Text):
                accumulator.append(item)
            else:
                self.visit(item)

                if len(accumulator) > 0:
                    new_items.append(
                        _Text(
                            content="".join(
                                text_element.content for text_element in accumulator
                            )
                        )
                    )
                    accumulator = []

                new_items.append(item)

        if len(accumulator) > 0:
            new_items.append(
                _Text(
                    content="".join(
                        text_element.content for text_element in accumulator
                    )
                )
            )

        node.items = new_items


class _RemoveRedundantParaVisitor(_NodeVisitor):
    """Remove the redundant ``<para>`` elements in-place."""

    def visit_element(self, node: _Element) -> None:
        self.visit(node.children)

        # noinspection PyUnresolvedReferences
        if (
            node.name in ("summary", "remarks", "li", "param", "returns", "para")
            and len(node.children.items) == 1
            and isinstance(node.children.items[0], _Element)
            and node.children.items[0].name == "para"
        ):
            # noinspection PyUnresolvedReferences
            node.children = node.children.items[0].children


def _compress_node_in_place(node: _NodeUnion) -> None:
    """Remove redundant nodes for more readability in the rendered text."""
    flatten_list_visitor = _FlattenListVisitor()
    flatten_list_visitor.visit(node)

    concatenate_text_visitor = _ConcatenateTextVisitor()
    concatenate_text_visitor.visit(node)

    remove_redundant_para_visitor = _RemoveRedundantParaVisitor()
    remove_redundant_para_visitor.visit(node)


def _render_summary_remarks(
    description: intermediate.SummaryRemarksDescription,
) -> Tuple[Optional[_List], Optional[List[Error]]]:
    """Render a description to our description node."""
    result_items = []  # type: List[_NodeUnion]
    errors = []  # type: List[Error]

    element_renderer = _ElementRenderer()

    summary_node, summary_errors = element_renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(
            Error(description.parsed.node, message) for message in summary_errors
        )
    else:
        assert summary_node is not None

        result_items.append(
            _Element(name="summary", children=_List(items=[summary_node]))
        )

    remark_nodes = []  # type: List[_NodeUnion]
    for remark in description.remarks:
        remark_node, remark_errors = element_renderer.transform(remark)
        if remark_errors:
            errors.extend(
                Error(description.parsed.node, message) for message in remark_errors
            )
        else:
            assert remark_node is not None
            remark_nodes.append(remark_node)

    if len(errors) > 0:
        return None, errors

    if len(remark_nodes) > 0:
        result_items.append(
            _Element(name="remarks", children=_List(items=remark_nodes))
        )

    return _List(items=result_items), None


def _render_summary_remarks_constraints(
    description: intermediate.SummaryRemarksConstraintsDescription,
) -> Tuple[Optional[_List], Optional[List[Error]]]:
    """Render a description where constraints are put in remarks."""
    result_items = []  # type: List[_NodeUnion]
    errors = []  # type: List[Error]

    element_renderer = _ElementRenderer()

    summary_node, summary_errors = element_renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(
            Error(description.parsed.node, message) for message in summary_errors
        )
    else:
        assert summary_node is not None
        result_items.append(
            _Element(name="summary", children=_List(items=[summary_node]))
        )

    remark_nodes = []  # type: List[_NodeUnion]
    for remark in description.remarks:
        remark_node, remark_errors = element_renderer.transform(remark)
        if remark_errors:
            errors.extend(
                Error(description.parsed.node, message) for message in remark_errors
            )
        else:
            assert remark_node is not None
            remark_nodes.append(remark_node)

    constraint_nodes = []  # type: List[_NodeUnion]
    for identifier, docutils_element in description.constraints_by_identifier.items():
        body, body_errors = element_renderer.transform(docutils_element)
        if body_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in body_errors
            )
        else:
            assert body is not None

            # NOTE (mristin, 2022-07-21):
            # We in-line the constraint prefix for better readability.

            # noinspection PyUnresolvedReferences
            if (
                isinstance(body, _List)
                and len(body.items) > 0
                and isinstance(body.items[0], _Element)
                and body.items[0].name == "para"
            ):
                # noinspection PyUnresolvedReferences
                body.items[0].children.items.insert(
                    0, _Text(content=f"Constraint {identifier}:\n")
                )

                constraint_node = _Element(name="li", children=body)
            else:
                constraint_node = _Element(
                    name="li",
                    children=_List(
                        items=[
                            _Element(
                                name="para",
                                children=_List(
                                    items=[_Text(content=f"Constraint {identifier}:\n")]
                                ),
                            ),
                            body,
                        ]
                    ),
                )

            constraint_nodes.append(constraint_node)

    if len(errors) > 0:
        return None, errors

    if len(constraint_nodes) > 0:
        remark_nodes.append(
            _Element(name="para", children=_List(items=[_Text(content="Constraints:")]))
        )

        ul_node = _Element(name="ul", children=_List(items=constraint_nodes))

        remark_nodes.append(ul_node)

    if len(remark_nodes) > 0:
        result_items.append(
            _Element(name="remarks", children=_List(items=remark_nodes))
        )

    return _List(items=result_items), None


@require(lambda line: "\n" not in line)
def _slash_slash_slash_line(line: str) -> str:
    """Prepend ``///`` to the ``line``."""
    if len(line) == 0:
        return "///"

    return f"/// {line}"


class _RelativeIndention:
    """
    Represent the relative indention.

    Since the indention is *relative*, it can be either positive, neutral or negative.
    """

    @require(lambda direction: direction in (-1, 0, 1))
    def __init__(self, direction: int) -> None:
        self.direction = direction

    def __repr__(self) -> str:
        """Generate text representation for easier debugging."""
        return f"{self.__class__.__name__}({self.direction!r})"


class _TextBlock(DBC):
    """
    Represent a block of text.

    This data structure is expected to be append-only mutable, where you keep adding
    new parts to the block. The parts are later expected to be joined by an empty
    string.

    All the text blocks are expected to be joined by empty strings.
    """

    def __init__(self, parts: List[str]) -> None:
        """Initialize with the given values."""
        self.parts = parts

    def __repr__(self) -> str:
        """Generate text representation for easier debugging."""
        return f"{self.__class__.__name__}({self.parts!r})"


class _EnforceNewLine(DBC):
    """
    Enforce that the following text starts on a new line.

    If there is already a new line output before, this text directive has no influence.
    """

    def __repr__(self) -> str:
        """Generate text representation for easier debugging."""
        return f"{self.__class__.__name__}()"


_TextDirective = Union[_RelativeIndention, _TextBlock, _EnforceNewLine]


class _ToTextDirectivesVisitor(_NodeVisitor):
    """
    Convert the nodes to a text as represented by text control directives.

    The text is expected to be valid XML and properly escaped.
    """

    #: The resulting text control directives
    directives: List[_TextDirective]

    def __init__(self) -> None:
        self.directives = []

    def _last_or_new_block(self) -> _TextBlock:
        """Retrieve the last block, or initialize a new block, if no last block."""
        if len(self.directives) == 0 or not isinstance(self.directives[-1], _TextBlock):
            self.directives.append(_TextBlock(parts=[]))

        assert isinstance(self.directives[-1], _TextBlock)
        return self.directives[-1]

    def visit_text(self, node: _Text) -> None:
        self._last_or_new_block().parts.append(xml.sax.saxutils.escape(node.content))

    def visit_element(self, node: _Element) -> None:
        """Visit the element node and its children."""
        if node.name in ("summary", "remarks", "para", "param", "returns"):
            # NOTE (mristin, 2022-07-18):
            # We render these tags without indention for better readability.

            start_element_writer = io.StringIO()
            start_element_writer.write(f"<{node.name}")
            if len(node.attrs) > 0:
                for attr_name, attr_value in node.attrs.items():
                    start_element_writer.write(
                        f" {attr_name}={xml.sax.saxutils.quoteattr(attr_value)}"
                    )
            start_element_writer.write(">")

            self.directives.append(_EnforceNewLine())

            self.directives.append(_TextBlock(parts=[start_element_writer.getvalue()]))

            self.directives.append(_EnforceNewLine())

            for item in node.children.items:
                self.visit(item)

            self.directives.append(_EnforceNewLine())

            self.directives.append(_TextBlock(parts=[f"</{node.name}>"]))

        elif node.name in ("ul", "li"):
            # NOTE (mristin, 2022-07-18):
            # We put the list elements on new lines and indent them.
            assert (
                len(node.attrs) == 0
            ), f"Unexpected attributes in a node {node.name!r}"

            self.directives.append(_EnforceNewLine())

            self.directives.append(_TextBlock(parts=[f"<{node.name}>"]))

            self.directives.append(_EnforceNewLine())

            self.directives.append(_RelativeIndention(direction=1))

            for item in node.children.items:
                self.visit(item)

            self.directives.append(_EnforceNewLine())

            self.directives.append(_RelativeIndention(direction=-1))

            self.directives.append(_TextBlock(parts=[f"</{node.name}>"]))

        else:
            # NOTE (mristin, 2022-07-18):
            # We inline all the other elements.

            start_element_writer = io.StringIO()
            start_element_writer.write(f"<{node.name}")
            if len(node.attrs) > 0:
                for attr_name, attr_value in node.attrs.items():
                    start_element_writer.write(
                        f" {attr_name}={xml.sax.saxutils.quoteattr(attr_value)}"
                    )

            if len(node.children.items) == 0:
                start_element_writer.write(" />")
                self._last_or_new_block().parts.append(start_element_writer.getvalue())

                return

            start_element_writer.write(">")
            self._last_or_new_block().parts.append(start_element_writer.getvalue())

            for item in node.children.items:
                self.visit(item)

            self._last_or_new_block().parts.append(f"</{node.name}>")


_TextDirectiveExceptEnforceNewLine = Union[_RelativeIndention, _TextBlock]
assert_union_without_excluded(
    original_union=_TextDirective,
    subset_union=_TextDirectiveExceptEnforceNewLine,
    excluded=[_EnforceNewLine],
)

T = TypeVar("T")


def pairwise(iterable: Iterable[T]) -> Iterator[Tuple[T, T]]:
    """Iterate pair-wise over the iterator."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


# fmt: off
@ensure(
    lambda result:
    all(
        len(directive.parts) > 0
        for directive in result
        if isinstance(directive, _TextBlock)
    ),
    "No empty text blocks"
)
@ensure(
    lambda result:
    all(
        not (
            isinstance(prev, _TextBlock)
            and isinstance(current, _TextBlock)
        )
        for prev, current in pairwise(result)
    ),
    "All text blocks are merged and there are no consecutive text blocks"
)
# fmt: on
def _compress_text_directives(
    directives: Sequence[_TextDirective],
) -> List[_TextDirectiveExceptEnforceNewLine]:
    """Merge consecutive text blocks and enforce the new lines."""
    # region Remove empty blocks

    directives_wo_empty_blocks = [
        directive
        for directive in directives
        if not (isinstance(directive, _TextBlock) and len(directive.parts) == 0)
    ]

    # endregion

    # region Fulfill new-line enforcement

    directives_wo_enforce_new_line = (
        []
    )  # type: List[_TextDirectiveExceptEnforceNewLine]

    previous_text_block = None  # type: Optional[_TextBlock]

    for directive in directives_wo_empty_blocks:
        if isinstance(directive, _EnforceNewLine):
            if previous_text_block is not None:
                assert len(previous_text_block.parts) > 0

                if not previous_text_block.parts[-1].endswith("\n"):
                    previous_text_block.parts.append("\n")
        elif isinstance(directive, _TextBlock):
            assert len(directive.parts) > 0

            previous_text_block = directive
            directives_wo_enforce_new_line.append(directive)

        elif isinstance(directive, _RelativeIndention):
            directives_wo_enforce_new_line.append(directive)
        else:
            assert_never(directive)

    # endregion

    # region Merge consecutive text blocks

    directives_w_merged_blocks = []  # type: List[_TextDirectiveExceptEnforceNewLine]

    for directive in directives_wo_enforce_new_line:
        if isinstance(directive, _TextBlock):
            assert len(directive.parts) > 0

            if len(directives_w_merged_blocks) > 0 and isinstance(
                directives_w_merged_blocks[-1], _TextBlock
            ):
                directives_w_merged_blocks[-1].parts.extend(directive.parts)
            else:
                directives_w_merged_blocks.append(directive)
        else:
            directives_w_merged_blocks.append(directive)

    # endregion

    return directives_w_merged_blocks


def _to_text(node: _NodeUnion) -> str:
    """
    Convert the node to a text representation.

    For readability and no phantom elements, the ``node`` is expected to be compressed
    before.
    """
    to_text_directives_visitor = _ToTextDirectivesVisitor()
    to_text_directives_visitor.visit(node)

    # NOTE (mristin, 2022-07-18):
    # We compress to do away with the new-line enforcement and consecutive and empty
    # blocks, so that the operations below become much easier to write.
    directives = _compress_text_directives(to_text_directives_visitor.directives)

    writer = io.StringIO()
    level = 0  # indention level

    for directive in directives:
        if isinstance(directive, _TextBlock):
            writer.write(textwrap.indent("".join(directive.parts), level * "  "))

        elif isinstance(directive, _RelativeIndention):
            assert level + directive.direction >= 0, (
                f"Negative absolute indention not possible: "
                f"{level=}, {directive.direction=}"
            )
            level += directive.direction

        else:
            assert_never(directive)

    return writer.getvalue()


def _generate_summary_remarks(
    description: intermediate.SummaryRemarksDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks description."""
    node, errors = _render_summary_remarks(description=description)
    if errors is not None:
        return None, errors

    assert node is not None

    _compress_node_in_place(node=node)
    text = _to_text(node)

    commented_lines = [_slash_slash_slash_line(line) for line in text.splitlines()]

    return Stripped("\n".join(commented_lines)), None


def _generate_summary_remarks_constraints(
    description: intermediate.SummaryRemarksConstraintsDescription,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks-constraints."""
    node, errors = _render_summary_remarks_constraints(description=description)
    if errors is not None:
        return None, errors

    assert node is not None

    _compress_node_in_place(node=node)
    text = _to_text(node)

    commented_lines = [_slash_slash_slash_line(line) for line in text.splitlines()]

    return Stripped("\n".join(commented_lines)), None


def generate_comment_for_meta_model(
    description: intermediate.DescriptionOfMetaModel,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given meta-model."""
    return _generate_summary_remarks_constraints(description)


def generate_comment_for_our_type(
    description: intermediate.DescriptionOfOurType,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for our type."""
    return _generate_summary_remarks_constraints(description)


def generate_comment_for_property(
    description: intermediate.DescriptionOfProperty,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given property."""
    return _generate_summary_remarks_constraints(description)


def generate_comment_for_enumeration_literal(
    description: intermediate.DescriptionOfEnumerationLiteral,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    return _generate_summary_remarks(description)


def _render_description_of_signature(
    description: intermediate.DescriptionOfSignature,
) -> Tuple[Optional[_List], Optional[List[Error]]]:
    """Render a description where constraints are put in remarks."""
    result_items = []  # type: List[_NodeUnion]
    errors = []  # type: List[Error]

    renderer = _ElementRenderer()

    summary_node, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(
            Error(description.parsed.node, message) for message in summary_errors
        )
    else:
        assert summary_node is not None
        result_items.append(
            _Element(name="summary", children=_List(items=[summary_node]))
        )

    remark_nodes = []  # type: List[_NodeUnion]
    for remark in description.remarks:
        remark_node, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(
                Error(description.parsed.node, message) for message in remark_errors
            )
        else:
            assert remark_node is not None
            remark_nodes.append(remark_node)

    param_nodes = []  # type: List[_NodeUnion]

    for name, docutils_element in description.arguments_by_name.items():
        returns, body_errors = renderer.transform(docutils_element)
        if body_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in body_errors
            )
        else:
            assert returns is not None

            param_nodes.append(
                _Element(
                    name="param",
                    attrs=collections.OrderedDict([("name", name)]),
                    children=_List(items=[returns]),
                )
            )

    returns_node = None  # type: Optional[_NodeUnion]

    if description.returns is not None:
        # NOTE (mristin, 2022-07-18):
        # We need to help the type checker in PyCharm a bit.
        assert isinstance(description.returns, docutils.nodes.field_body)

        returns, returns_errors = renderer.transform(description.returns)
        if returns_errors is not None:
            errors.extend(
                Error(description.parsed.node, message) for message in returns_errors
            )
        else:
            assert returns is not None

            returns_node = _Element(name="returns", children=_List(items=[returns]))

    if len(errors) > 0:
        return None, errors

    if len(remark_nodes) > 0:
        result_items.append(
            _Element(name="remarks", children=_List(items=remark_nodes))
        )

    result_items.extend(param_nodes)

    if returns_node is not None:
        result_items.append(returns_node)

    return _List(items=result_items), None


def generate_comment_for_signature(
    description: intermediate.DescriptionOfSignature,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """
    Generate the documentation comment for the given signature.

    A signature, in this context, means a function or a method signature.
    """
    node, errors = _render_description_of_signature(description=description)
    if errors is not None:
        return None, errors

    assert node is not None

    _compress_node_in_place(node=node)
    text = _to_text(node)

    commented_lines = [_slash_slash_slash_line(line) for line in text.splitlines()]

    return Stripped("\n".join(commented_lines)), None



"""Render descriptions to documentation comments."""
import io
import textwrap
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Final,
    Union,
)

import docutils.nodes
import docutils.utils

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Error,
    assert_never,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.intermediate import (
    doc as intermediate_doc,
)
from aas_core_codegen.python import (
    common as python_common,
    naming as python_naming,
)


class Context:
    """Structure the context of the rendering of a description."""

    #: Qualified module name of the base module for the SDK
    qualified_module_name: Final[python_common.QualifiedModuleName]

    #: Our internal module in which the description resides
    module: Final[Identifier]

    #: Scope in which the description resides.
    #:
    #: This narrows down the outer scope of the description, while the description
    #: does not necessarily describe a class, but can also describe a method,
    #: a property, an enumeration literal *etc.*
    #:
    cls_or_enum: Final[
        Optional[Union[intermediate.ClassUnion, intermediate.Enumeration]]
    ]

    def __init__(
        self,
        qualified_module_name: python_common.QualifiedModuleName,
        module: Identifier,
        cls_or_enum: Optional[Union[intermediate.ClassUnion, intermediate.Enumeration]],
    ) -> None:
        """Initialize with the given values."""
        self.qualified_module_name = qualified_module_name
        self.module = module
        self.cls_or_enum = cls_or_enum


class _ElementRenderer(intermediate_doc.DocutilsElementTransformer[str]):
    """Render descriptions as a content of a docstring."""

    def __init__(self, context: Context) -> None:
        """Initialize with the given values."""
        self.context = context

    def transform_text(
        self, element: docutils.nodes.Text
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return element.astext(), None

    def transform_reference_to_our_type_in_doc(
        self, element: intermediate_doc.ReferenceToOurType
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        result: str

        name = python_naming.class_name(element.our_type.name)

        if isinstance(
            element.our_type,
            (
                intermediate.Enumeration,
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            if self.context.module == "types":
                result = f":py:class:`{name}`"
            else:
                result = f":py:class:`.types.{name}`"

        elif isinstance(element.our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-09-08):
            # We do not generate a class for constrained primitives, but we
            # leave it here as a literal.
            result = f"``{name}``"

        else:
            assert_never(element.our_type)

        return result, None

    def transform_reference_to_attribute_in_doc(
        self, element: intermediate_doc.ReferenceToAttribute
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        result: str

        if isinstance(element.reference, intermediate_doc.ReferenceToProperty):
            cls_name = python_naming.class_name(element.reference.cls.name)
            prop_name = python_naming.property_name(element.reference.prop.name)

            if self.context.module == "types":
                if self.context.cls_or_enum is element.reference.cls:
                    result = f":py:attr:`{prop_name}`"
                else:
                    result = f":py:attr:`{cls_name}.{prop_name}`"
            else:
                result = f":py:attr:`.types.{cls_name}.{prop_name}`"

        elif isinstance(
            element.reference, intermediate_doc.ReferenceToEnumerationLiteral
        ):
            cls_name = python_naming.class_name(element.reference.enumeration.name)
            literal_name = python_naming.enum_literal_name(
                element.reference.literal.name
            )

            if self.context.module == "types":
                if self.context.cls_or_enum is element.reference.enumeration:
                    result = f":py:attr:`{literal_name}`"
                else:
                    result = f":py:attr:`{cls_name}.{literal_name}`"
            else:
                result = f":py:attr:`.types.{cls_name}.{literal_name}`"
        else:
            assert_never(element.reference)

        return result, None

    def transform_reference_to_argument_in_doc(
        self, element: intermediate_doc.ReferenceToArgument
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        # NOTE (mristin, 2022-09-08):
        # We rely here on sphinx-paramlinks extension.

        arg_name = python_naming.argument_name(Identifier(element.reference))

        return f":paramref:`{arg_name}`", None

    def transform_reference_to_constraint_in_doc(
        self, element: intermediate_doc.ReferenceToConstraint
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        assert isinstance(element.reference, str)

        return (
            f":ref:`Constraint {element.reference} <constraint_{element.reference}>`",
            None,
        )

    def transform_reference_to_constant_in_doc(
        self, element: intermediate_doc.ReferenceToConstant
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        name = python_naming.constant_name(element.constant.name)

        if self.context.module == "constants":
            result = f":py:attr:`{name}`"
        else:
            result = f":py:attr:`.constants.{name}`"

        return result, None

    def transform_literal(
        self, element: docutils.nodes.literal
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        # NOTE (mristin, 2022-09-08):
        # We fail here catastrophically if there are backticks as there is no easy way
        # to escape them in RST. However, since our meta-model is also written in
        # Python, this assertion will almost always pass.
        #
        # See: https://stackoverflow.com/questions/66435475/how-to-escape-the-backtick-character-in-a-rst-file
        text = element.astext()
        assert "`" not in text
        return f"``{text}``", None

    def _render_children_concatenated_with_paragraph_breaks_where_necessary(
        self, children: Sequence[docutils.nodes.Element]
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        """Render the children and join them by empty lines, where appropriate."""
        writer = io.StringIO()
        previous_child = None  # type: Optional[docutils.nodes.Element]

        errors = []  # type: List[str]

        for child in children:
            text, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
                continue
            else:
                assert text is not None
                if isinstance(
                    previous_child,
                    (
                        docutils.nodes.paragraph,
                        docutils.nodes.bullet_list,
                        docutils.nodes.note,
                    ),
                ):
                    writer.write("\n\n")

                writer.write(text)

            previous_child = child

        if len(errors) > 0:
            return None, errors

        return writer.getvalue(), None

    def transform_paragraph(
        self, element: docutils.nodes.paragraph
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return self._render_children_concatenated_with_paragraph_breaks_where_necessary(
            element.children
        )

    def transform_emphasis(
        self, element: docutils.nodes.emphasis
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        children = []  # type: List[str]
        errors = []  # type: List[str]
        for child in element.children:
            rendered_child, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
            else:
                assert rendered_child is not None
                children.append(rendered_child)

        if len(errors) > 0:
            return None, errors

        content = "".join(children)
        return f"*{content}*", None

    def transform_list_item(
        self, element: docutils.nodes.list_item
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return self._render_children_concatenated_with_paragraph_breaks_where_necessary(
            element.children
        )

    def transform_bullet_list(
        self, element: docutils.nodes.bullet_list
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        errors = []  # type: List[str]
        writer = io.StringIO()

        for i, child in enumerate(element.children):
            assert isinstance(child, docutils.nodes.list_item), (
                f"Expected a list item in the bullet list, "
                f"but got an instance of {type(child)}: {child}"
            )

            child_text, child_errors = self.transform(child)
            if child_errors is not None:
                errors.extend(child_errors)
                continue

            assert child_text is not None
            if i > 0:
                writer.write("\n")

            writer.write("* ")

            # NOTE (mristin, 2022-09-14):
            # This has a potentially exponential complexity w.r.t. indention level.
            # However, as the indention level is thus far limited to only a single
            # level, we ignore this pitfall for the moment.
            #
            # In the future, if the indention level increases, we need to 1) change
            # the logic anyhow to allow for nested bullet lists, and 2) use data
            # structures that allow for more dynamic handling of the indention.
            writer.write(indent_but_first_line(child_text, "  "))

        if len(errors) > 0:
            return None, errors

        return writer.getvalue(), None

    def transform_note(
        self, element: docutils.nodes.note
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        writer = io.StringIO()
        writer.write(".. note::\n\n")

        (
            text,
            errors,
        ) = self._render_children_concatenated_with_paragraph_breaks_where_necessary(
            element.children
        )
        if errors is not None:
            return None, errors

        assert text is not None
        writer.write(textwrap.indent(text, "    "))
        return writer.getvalue(), None

    def transform_reference(
        self, element: docutils.nodes.reference
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return element.astext(), None

    def transform_field_body(
        self, element: docutils.nodes.field_body
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return self._render_children_concatenated_with_paragraph_breaks_where_necessary(
            element.children
        )

    def transform_document(
        self, element: docutils.nodes.document
    ) -> Tuple[Optional[str], Optional[List[str]]]:
        return self._render_children_concatenated_with_paragraph_breaks_where_necessary(
            element.children
        )


def generate_summary_remarks(
    description: intermediate.SummaryRemarksDescription, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks."""
    errors = []  # type: List[str]

    blocks = []  # type: List[str]

    renderer = _ElementRenderer(context=context)

    rendered_summary, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(summary_errors)
    else:
        assert rendered_summary is not None
        blocks.append(rendered_summary)

    for remark in description.remarks:
        rendered_remark, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(remark_errors)
        else:
            assert rendered_remark is not None
            blocks.append(rendered_remark)

    if len(errors) > 0:
        return None, [
            Error(description.parsed.node, error_message) for error_message in errors
        ]

    return Stripped("\n\n".join(blocks)), None


def generate_summary_remarks_constraints(
    description: intermediate.SummaryRemarksConstraintsDescription, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for a summary-remarks-constraints."""
    errors = []  # type: List[str]

    blocks = []  # type: List[str]

    renderer = _ElementRenderer(context=context)

    rendered_summary, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(summary_errors)
    else:
        assert rendered_summary is not None
        blocks.append(rendered_summary)

    for remark in description.remarks:
        rendered_remark, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(remark_errors)
        else:
            assert rendered_remark is not None
            blocks.append(rendered_remark)

    for constraint_id, constraint in description.constraints_by_identifier.items():
        constraint_body, constraint_errors = renderer.transform(constraint)
        if constraint_errors is not None:
            errors.extend(constraint_errors)
        else:
            assert constraint_body is not None
            writer = io.StringIO()
            writer.write(f":constraint {constraint_id}:\n")
            writer.write(f"    .. _constraint_{constraint_id}:\n\n")
            writer.write(textwrap.indent(constraint_body, "    "))
            blocks.append(writer.getvalue())

    if len(errors) > 0:
        return None, [
            Error(description.parsed.node, error_message) for error_message in errors
        ]

    return Stripped("\n\n".join(blocks)), None


def docstring(text: Stripped) -> Stripped:
    """Generate a docstring out of the text."""
    escaped = text.replace("\\", "\\\\").replace('"""', '\\"\\"\\"')
    if 3 + len(escaped) + 3 < 70:
        return Stripped(f'"""{escaped}"""')

    return Stripped(f'"""\n{escaped}\n"""')


def documentation_comment(text: Stripped) -> Stripped:
    """Generate the documentation comment with the given ``text``."""
    commented_lines = []  # type: List[str]
    for line in text.splitlines():
        if len(line.strip()) == 0:
            commented_lines.append("#:")
        else:
            commented_lines.append(f"#: {line}")

    return Stripped("\n".join(commented_lines))


def generate_docstring_for_signature(
    description: intermediate.DescriptionOfSignature, context: Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """
    Generate the docstring for the given signature.

    A signature, in this context, means a function or a method signature.
    """
    errors = []  # type: List[str]

    blocks = []  # type: List[str]

    renderer = _ElementRenderer(context=context)

    rendered_summary, summary_errors = renderer.transform(description.summary)
    if summary_errors is not None:
        errors.extend(summary_errors)
    else:
        assert rendered_summary is not None
        blocks.append(rendered_summary)

    for remark in description.remarks:
        rendered_remark, remark_errors = renderer.transform(remark)
        if remark_errors:
            errors.extend(remark_errors)
        else:
            assert rendered_remark is not None
            blocks.append(rendered_remark)

    param_and_return_blocks = []  # type: List[Stripped]
    for arg_name, arg_description in description.arguments_by_name.items():
        rendered_arg_description, arg_errors = renderer.transform(arg_description)
        if arg_errors is not None:
            errors.extend(arg_errors)
            continue

        assert rendered_arg_description is not None

        if "\n" not in rendered_arg_description and (
            7 + len(arg_name) + 1 + len(rendered_arg_description) < 60
        ):
            param_and_return_blocks.append(
                Stripped(f":param {arg_name}: {rendered_arg_description}")
            )
        else:
            writer = io.StringIO()
            writer.write(f":param {arg_name}:\n")
            writer.write(textwrap.indent(rendered_arg_description, "    "))
            param_and_return_blocks.append(Stripped(writer.getvalue()))

    if description.returns is not None:
        rendered_returns, returns_errors = renderer.transform(description.returns)
        if returns_errors is not None:
            errors.extend(returns_errors)
        else:
            assert rendered_returns is not None

            if "\n" not in rendered_returns and (9 + len(rendered_returns) < 60):
                param_and_return_blocks.append(Stripped(f":return: {rendered_returns}"))
            else:
                writer = io.StringIO()
                writer.write(":return:\n")
                writer.write(textwrap.indent(rendered_returns, "    "))
                param_and_return_blocks.append(Stripped(writer.getvalue()))

    if len(param_and_return_blocks) > 0:
        blocks.append("\n".join(param_and_return_blocks))

    if len(errors) > 0:
        return None, [
            Error(description.parsed.node, error_message) for error_message in errors
        ]

    text = Stripped("\n\n".join(blocks))

    return docstring(text), None



"""Generate JSON schema corresponding to the meta-model."""
import collections
import json
from typing import (
    TextIO,
    Any,
    MutableMapping,
    Optional,
    Tuple,
    List,
    Sequence,
    Mapping,
    Callable,
)

from icontract import ensure, require

import aas_core_codegen.jsonschema
from aas_core_codegen import (
    naming,
    specific_implementations,
    intermediate,
    run,
    infer_for_schema,
)
from aas_core_codegen.parse import retree as parse_retree
from aas_core_codegen.common import Stripped, Error, assert_never, Identifier

assert aas_core_codegen.jsonschema.__doc__ == __doc__


def _define_for_enumeration(
    enumeration: intermediate.Enumeration,
) -> MutableMapping[str, Any]:
    """
    Generate the definition for an ``enumeration``.

    The list of definitions is to be *extended* with the resulting mapping.
    """
    definition = collections.OrderedDict()  # type: MutableMapping[str, Any]
    definition["type"] = "string"
    definition["enum"] = sorted(literal.value for literal in enumeration.literals)

    model_type = naming.json_model_type(enumeration.name)

    return collections.OrderedDict([(model_type, definition)])


_PRIMITIVE_MAP = {
    intermediate.PrimitiveType.BOOL: "boolean",
    intermediate.PrimitiveType.INT: "integer",
    intermediate.PrimitiveType.FLOAT: "number",
    intermediate.PrimitiveType.STR: "string",
    intermediate.PrimitiveType.BYTEARRAY: "string",
}
assert all(literal in _PRIMITIVE_MAP for literal in intermediate.PrimitiveType)


def _define_primitive_type(
    primitive_type: intermediate.PrimitiveType,
) -> MutableMapping[str, Any]:
    """Generate the definition for the given ``primitive_type``."""
    definition = collections.OrderedDict([("type", _PRIMITIVE_MAP[primitive_type])])

    if primitive_type is intermediate.PrimitiveType.BYTEARRAY:
        definition["contentEncoding"] = "base64"

    return definition


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_type(
    type_annotation: intermediate.TypeAnnotationExceptOptional,
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[Error]]:
    """Generate the type definition for ``type_annotation``."""
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return _define_primitive_type(type_annotation.a_type), None

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        model_type = naming.json_model_type(type_annotation.our_type.name)

        if isinstance(type_annotation.our_type, intermediate.Enumeration):
            return (
                collections.OrderedDict([("$ref", f"#/definitions/{model_type}")]),
                None,
            )

        elif isinstance(type_annotation.our_type, intermediate.ConstrainedPrimitive):
            return _define_primitive_type(type_annotation.our_type.constrainee), None

        elif isinstance(type_annotation.our_type, intermediate.Class):
            if len(type_annotation.our_type.concrete_descendants) > 0:
                return (
                    collections.OrderedDict(
                        [("$ref", f"#/definitions/{model_type}_choice")]
                    ),
                    None,
                )
            else:
                return (
                    collections.OrderedDict([("$ref", f"#/definitions/{model_type}")]),
                    None,
                )

        else:
            assert_never(type_annotation.our_type)

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        assert not isinstance(
            type_annotation.items, intermediate.OptionalTypeAnnotation
        ), (
            "NOTE (mristin, 2023-02-06): Lists of optional values were not expected "
            "at the time when we implemented this. Please contact the developers "
            "if you need this functionality."
        )

        items_type_definition, items_error = _define_type(
            type_annotation=type_annotation.items
        )

        if items_error is not None:
            return None, items_error

        assert items_type_definition is not None

        return (
            collections.OrderedDict(
                [("type", "array"), ("items", items_type_definition)]
            ),
            None,
        )

    else:
        raise NotImplementedError(
            f"(mristin, 2021-11-10):\n"
            f"We implemented only a subset of possible type annotations "
            f"to be represented in a JSON schema since we lacked more information "
            f"about the context.\n\n"
            f"This feature needs yet to be implemented.\n\n"
            f"{type_annotation=}"
        )


# NOTE (mristin):
# This function is made public so that we can use it in other schema generators such
# as the SHACL generator.
def fix_pattern_for_utf16(pattern: str) -> str:
    """Parse the pattern and re-render it for UTF-16-only regex engines."""
    regex, error = parse_retree.parse([pattern])
    if error is not None:
        raise ValueError(
            f"The pattern could not be parsed: {pattern!r}; error was: {error}"
        )

    assert regex is not None
    parse_retree.fix_for_utf16_regex_in_place(regex)

    parts = parse_retree.render(regex=regex)
    assert all(
        isinstance(part, str) for part in parts
    ), "Only string parts expected, no formatted values"

    # NOTE (mristin, 2023-03-15):
    # We have to make this transformation for mypy.
    parts_str = []  # type: List[str]
    for part in parts:
        assert isinstance(part, str), "Only string parts expected, no formatted values"
        parts_str.append(part)

    return "".join(parts_str)


def _define_constraints_for_primitive_type(
    primitive_type: intermediate.PrimitiveType,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
    fix_pattern: Callable[[str], str],
) -> MutableMapping[str, Any]:
    """
    Generate the constraints, if any, for the ``primitive_type``.

    If there are no constraints, an empty mapping is returned.

    The ``fix_pattern`` determines how the pattern should be translated for
    the regex engine. For example, some JSON schema verification engines expect only
    characters below Basic Multilingual Plane (BMP), and use surrogate pairs to
    represent characters above BMP.
    """
    definition = collections.OrderedDict()  # type: MutableMapping[str, Any]

    if (
        primitive_type
        in (intermediate.PrimitiveType.STR, intermediate.PrimitiveType.BYTEARRAY)
        and len_constraint is not None
    ):
        if len_constraint.min_value is not None:
            definition["minLength"] = len_constraint.min_value

        if len_constraint.max_value is not None:
            definition["maxLength"] = len_constraint.max_value

    if (
        primitive_type == intermediate.PrimitiveType.STR
        and pattern_constraints is not None
        and len(pattern_constraints) > 0
    ):
        if len(pattern_constraints) == 1:
            definition["pattern"] = fix_pattern(pattern_constraints[0].pattern)
        else:
            all_of = [definition]  # type: List[MutableMapping[str, Any]]

            for pattern_constraint in pattern_constraints:
                all_of.append(
                    collections.OrderedDict(
                        [
                            (
                                "pattern",
                                fix_pattern(pattern_constraint.pattern),
                            )
                        ]
                    )
                )

            definition = collections.OrderedDict([("allOf", all_of)])

    return definition


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_constraints(
    type_annotation: intermediate.TypeAnnotation,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
    fix_pattern: Callable[[str], str],
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[Error]]:
    """
    Generate only the constraints for the given ``type_annotation``.

    The type is generated in a separated function, `_generate_type`.

    The ``fix_pattern`` determines how the pattern should be translated for
    the regex engine. For example, some JSON schema verification engines expect only
    characters below Basic Multilingual Plane (BMP), and use surrogate pairs to
    represent characters above BMP.

    If there are no constraints, an empty mapping is returned.
    """
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return (
            _define_constraints_for_primitive_type(
                primitive_type=type_annotation.a_type,
                len_constraint=len_constraint,
                pattern_constraints=pattern_constraints,
                fix_pattern=fix_pattern,
            ),
            None,
        )

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        if isinstance(type_annotation.our_type, intermediate.Enumeration):
            return collections.OrderedDict(), None

        elif isinstance(type_annotation.our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-02-11):
            # We in-line the constraints from the constrained primitives directly
            # in the properties. We do not want to introduce separate definitions
            # for them as that would make it more difficult for downstream code
            # generators to generate meaningful code (*e.g.*, code generators for
            # OpenAPI3).
            return (
                _define_constraints_for_primitive_type(
                    primitive_type=type_annotation.our_type.constrainee,
                    len_constraint=len_constraint,
                    pattern_constraints=pattern_constraints,
                    fix_pattern=fix_pattern,
                ),
                None,
            )

        elif isinstance(type_annotation.our_type, intermediate.Class):
            # NOTE (mristin, 2023-02-06):
            # There are no constraints for class itself in JSON schema. We define
            # constraints only per properties at the moment.
            return collections.OrderedDict(), None

        else:
            assert_never(type_annotation.our_type)

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        # NOTE (mristin, 2021-12-02):
        # We do not propagate the inference of constraints to sub-lists
        # so in this case we set all the constraint on the ``items`` to none.
        # This behavior might change in the future when we encounter such
        # constraints and have more context available.

        definition = collections.OrderedDict()

        if len_constraint is not None:
            if len_constraint.min_value is not None:
                assert (
                    "minItems" not in definition
                ), "Unexpected property 'minItems' in the JSON type definition"

                definition["minItems"] = len_constraint.min_value

            if len_constraint.max_value is not None:
                assert (
                    "maxItems" not in definition
                ), "Unexpected property 'maxItems' in the JSON type definition"

                definition["maxItems"] = len_constraint.max_value

        return definition, None

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        raise NotImplementedError(
            f"(mristin, 2021-11-10):\n"
            f"Nested optional values are unexpected in the JSON schema. "
            f"We did not implement them at the moment since we need more information "
            f"about the context.\n\n"
            f"This feature needs yet to be implemented.\n\n"
            f"{type_annotation=}"
        )

    else:
        raise NotImplementedError(
            f"(mristin, 2021-11-10):\n"
            f"We implemented only a subset of possible type annotations "
            f"to be represented in a JSON schema since we lacked more information "
            f"about the context.\n\n"
            f"This feature needs yet to be implemented.\n\n"
            f"{type_annotation=}"
        )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_properties(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
    fix_pattern: Callable[[str], str],
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[List[Error]]]:
    """
    Generate the definitions of the meta-model properties for the given ``cls``.

    The property ``modelType`` is defined separately as we need to distinguish
    cases where it is set (concrete class) and not (abstract class, or stub of
    a concrete class with descendants).

    The ``fix_pattern`` determines how the pattern should be translated for
    the regex engine. For example, some JSON schema verification engines expect only
    characters below Basic Multilingual Plane (BMP), and use surrogate pairs to
    represent characters above BMP.
    """
    errors = []  # type: List[Error]

    properties = collections.OrderedDict()  # type: MutableMapping[str, Any]

    for prop in cls.properties:
        prop_name = naming.json_property(prop.name)

        len_constraint = constraints_by_property.len_constraints_by_property.get(
            prop, None
        )

        pattern_constraints = constraints_by_property.patterns_by_property.get(
            prop, None
        )

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        definition = collections.OrderedDict()  # type: MutableMapping[str, Any]

        # NOTE (mristin, 2023-02-06):
        # The properties are inherited through ``allOf``. We can not change the type
        # of property in the children as this would result in a conflict and
        # unsatisfiable schema. Therefore, we define the type for a property only
        # at the parent class, where the property is defined for the first time.
        if prop.specified_for is cls:
            property_definition, error = _define_type(type_annotation=type_anno)

            if error is not None:
                errors.append(error)
            else:
                assert property_definition is not None
                definition.update(property_definition)

        constraints_definition, error = _define_constraints(
            type_annotation=type_anno,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
            fix_pattern=fix_pattern,
        )
        if error is not None:
            errors.append(error)
        else:
            assert constraints_definition is not None
            definition.update(constraints_definition)

        # NOTE (mristin, 2023-02-06):
        # We do not want to pollute the schema with empty definitions. An empty
        # definition results if there are no constraints and the property is inherited
        # from an ancestor class through ``allOf``.
        if len(definition) > 0:
            properties[prop_name] = definition

    if len(errors) > 0:
        return None, errors

    return properties, None


def _list_required_properties(cls: intermediate.ClassUnion) -> List[Identifier]:
    """
    List all the properties which are required.

    Only the meta-model properties are listed, so this list might not be exhaustive.
    For example, the JSON property ``modelType`` is not listed here.
    """
    required = []  # type: List[Identifier]
    for prop in cls.properties:
        # NOTE (mristin, 2023-02-06):
        # We stack the inheritance as ``allOf``. This will impose the stacking
        # of the required fields as well, so whenever you add a field to
        # a child ``required`` constraint, it will *extend* the list of
        # the required fields, *not* replace it.
        if prop.specified_for is not cls:
            continue

        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            prop_name = naming.json_property(prop.name)
            required.append(prop_name)

    return required


def _define_all_of_for_inheritance(
    cls: intermediate.ClassUnion,
) -> List[MutableMapping[str, Any]]:
    """Generate ``allOf`` definition as inheritance."""
    all_of = []  # type: List[MutableMapping[str, Any]]

    for inheritance in cls.inheritances:
        if isinstance(inheritance, intermediate.AbstractClass):
            all_of.append(
                {"$ref": f"#/definitions/{naming.json_model_type(inheritance.name)}"}
            )
        elif isinstance(inheritance, intermediate.ConcreteClass):
            # NOTE (mristin, 2023-03-13):
            # We distinguish between two definitions corresponding to the same concrete
            # class:
            #
            # 1) One definition defines only the class in abstract, so that
            #    it can be inherited. The abstract definition lacks the ``modelType``
            #    constant, as that would conflict in inheritance.
            # 2) The other definition corresponds to the concrete definition of
            #    the class that an instance has to fulfill. This definition includes
            #    the constant ``modelType``.
            #
            # We had to separate these two definitions to avoid conflicts in
            # ``modelType`` constant between a parent concrete class and a child
            # concrete class.
            all_of.append(
                {
                    "$ref": f"#/definitions/"
                    f"{naming.json_model_type(inheritance.name)}_abstract"
                }
            )

    return all_of


@require(lambda cls: len(cls.concrete_descendants) > 0)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_inheritable_definition(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
    fix_pattern: Callable[[str], str],
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[List[Error]]]:
    """
    Generate a definition of ``cls`` for inheritance through ``allOf``.

    The ``fix_pattern`` determines how the pattern should be translated for
    the respective JSON schema engine.

    The definitions are to be *extended* with the resulting mapping.
    """
    all_of = _define_all_of_for_inheritance(cls)

    errors = []  # type: List[Error]

    properties, properties_error = _define_properties(
        cls=cls,
        constraints_by_property=constraints_by_property,
        fix_pattern=fix_pattern,
    )
    if properties_error is not None:
        errors.extend(properties_error)

    if len(errors) > 0:
        return None, errors

    assert properties is not None

    required = _list_required_properties(cls)

    if cls.serialization.with_model_type and not any(
        inheritance.serialization.with_model_type for inheritance in cls.inheritances
    ):
        # NOTE (mristin, 2023-03-13):
        # This is going to be an abstract definition for inheritance, so we can not pin
        # the ``modelType`` to a fixed, constant value.
        assert "modelType" not in properties
        properties["modelType"] = {"$ref": "#/definitions/ModelType"}

        required.append(Identifier("modelType"))

    definition = collections.OrderedDict()  # type: MutableMapping[str, Any]
    if len(cls.inheritances) == 0:
        definition["type"] = "object"

    if len(properties) > 0:
        definition["properties"] = properties

        if len(required) > 0:
            definition["required"] = required

    if len(definition) > 0:
        all_of.append(definition)

    definition_name: str
    if isinstance(cls, intermediate.AbstractClass):
        definition_name = naming.json_model_type(cls.name)
    elif isinstance(cls, intermediate.ConcreteClass):
        definition_name = f"{naming.json_model_type(cls.name)}_abstract"
    else:
        assert_never(cls)

    result = collections.OrderedDict()  # type: MutableMapping[str, Any]

    if len(all_of) == 0:
        result[definition_name] = {"type": "object"}
    elif len(all_of) == 1:
        result[definition_name] = all_of[0]
    else:
        result[definition_name] = {"allOf": all_of}

    return result, None


@require(lambda cls: len(cls.concrete_descendants) > 0)
def _generate_choice_definition(
    cls: intermediate.ClassUnion,
) -> MutableMapping[str, Any]:
    """
    Generate the definition of dispatching through ``oneOf``.

    The definitions are to be *extended* with the resulting mapping.
    """
    one_of = []  # type: List[Mapping[str, Any]]
    if isinstance(cls, intermediate.ConcreteClass):
        one_of.append({"$ref": f"#/definitions/{naming.json_model_type(cls.name)}"})

    for descendant in cls.concrete_descendants:
        one_of.append(
            {"$ref": f"#/definitions/{naming.json_model_type(descendant.name)}"}
        )

    return {f"{naming.json_model_type(cls.name)}_choice": {"oneOf": one_of}}


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_concrete_definition(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
    fix_pattern: Callable[[str], str],
) -> Tuple[Optional[MutableMapping[str, Any]], Optional[List[Error]]]:
    """
    Generate the definition of a concrete class to be matched by an instance.

    The ``fix_pattern`` determines how the pattern should be translated for
    the respective JSON schema engine.
    """
    # NOTE (mristin, 2023-03-13):
    # We distinguish between two definitions corresponding to the same concrete
    # class:
    #
    # 1) One definition defines only the class in abstract, so that
    #    it can be inherited. The abstract definition lacks the ``modelType``
    #    constant, as that would conflict in inheritance.
    # 2) The other definition corresponds to the concrete definition of
    #    the class that an instance has to fulfill. This definition includes
    #    the constant ``modelType``.
    #
    # We had to separate these two definitions to avoid conflicts in
    # ``modelType`` constant between a parent concrete class and a child
    # concrete class.

    model_type = naming.json_model_type(cls.name)

    if len(cls.concrete_descendants) > 0:
        assert cls.serialization.with_model_type, (
            f"Expected model type to be included in a class with concrete descendants "
            f"{cls.name!r}"
        )

        all_of = [
            {"$ref": f"#/definitions/{model_type}_abstract"},
            {"properties": {"modelType": {"const": model_type}}},
        ]  # type: List[MutableMapping[str, Any]]

        return {model_type: {"allOf": all_of}}, None

    all_of = _define_all_of_for_inheritance(cls)

    errors = []  # type: List[Error]

    properties, properties_error = _define_properties(
        cls=cls,
        constraints_by_property=constraints_by_property,
        fix_pattern=fix_pattern,
    )
    if properties_error is not None:
        errors.extend(properties_error)

    if len(errors) > 0:
        return None, errors

    assert properties is not None

    required = _list_required_properties(cls)

    if cls.serialization.with_model_type:
        properties["modelType"] = {"const": model_type}

    definition = collections.OrderedDict()  # type: MutableMapping[str, Any]
    if len(cls.inheritances) == 0:
        definition["type"] = "object"

    if len(properties) > 0:
        definition["properties"] = properties

        if len(required) > 0:
            definition["required"] = required

    all_of.append(definition)

    model_type = naming.json_model_type(cls.name)

    result = collections.OrderedDict()  # type: MutableMapping[str, Any]

    if len(all_of) == 0:
        result[model_type] = {"type": "object"}
    elif len(all_of) == 1:
        result[model_type] = all_of[0]
    else:
        result[model_type] = {"allOf": all_of}

    return result, None


class Definitions:
    """Store definitions of the schema as we go."""

    def __init__(self) -> None:
        """Initialize as empty."""
        self._definitions = collections.OrderedDict()  # type: MutableMapping[str, Any]

    def get(self) -> Mapping[str, Any]:
        """Get the content."""
        return self._definitions

    def update_for(
        self, our_type: intermediate.OurType, extension: Mapping[str, Any]
    ) -> Optional[Error]:
        """Update the definitions with ``extension`` related to ``our_type``."""
        for key, definition in extension.items():
            if key in self._definitions:
                return Error(
                    our_type.parsed.node,
                    f"One of the JSON definitions, {key}, "
                    f"for our type {our_type.name} has been "
                    f"already provided in the definitions; "
                    f"did you already perhaps define it in another "
                    f"implementation-specific snippet?",
                )

            self._definitions[key] = definition

        return None

    def update(self, extension: Mapping[str, Any]) -> Optional[Error]:
        """Update the definitions with ``extension`` unrelated to any of our types."""
        for key, definition in extension.items():
            if key in self._definitions:
                return Error(
                    None,
                    f"One of the JSON definitions, {key}, "
                    f"has been already provided in the definitions; "
                    f"did you already perhaps define it in another "
                    f"implementation-specific snippet?",
                )

            self._definitions[key] = definition

        return None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    fix_pattern: Callable[[str], str],
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """
    Generate the JSON schema based on the symbol table.

    The ``fix_pattern`` determines how the pattern should be translated for
    the respective JSON schema engine.

    This function is intended to be used not only by aas-core-codegen, but also for
    downstream clients. For example, the downstream clients should use this function
    to customize how patterns should be rendered / fixed for the respective regex
    engine.
    """
    schema_base_key = specific_implementations.ImplementationKey("schema_base.json")

    schema_base_json = spec_impls.get(schema_base_key, None)
    if schema_base_json is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the base schema "
                f"is missing: {schema_base_key}",
            )
        ]

    schema: MutableMapping[str, Any]

    try:
        # noinspection PyTypeChecker
        schema = json.loads(schema_base_json, object_pairs_hook=collections.OrderedDict)
    except json.JSONDecodeError as err:
        return None, [
            Error(
                None, f"Failed to parse the base schema from {schema_base_key}: {err}"
            )
        ]

    if "$id" in schema:
        return None, [
            Error(
                None,
                f"Unexpected property '$id' in the base JSON schema "
                f"from: {schema_base_key}",
            )
        ]

    # NOTE (mristin, 2022-08-25):
    # We use the same namespace in all the schemas for the consistency.
    schema["$id"] = symbol_table.meta_model.xml_namespace

    if "definitions" in schema:
        return None, [
            Error(
                None,
                f"The property ``definitions`` unexpected in the base JSON schema "
                f"from: {schema_base_key}",
            )
        ]

    errors = []  # type: List[Error]

    definitions = Definitions()

    constraints_by_class, some_errors = infer_for_schema.infer_constraints_by_class(
        symbol_table=symbol_table
    )

    if some_errors is not None:
        errors.extend(some_errors)

    if len(errors) > 0:
        return None, errors

    assert constraints_by_class is not None

    ids_of_our_types_in_properties = (
        intermediate.collect_ids_of_our_types_in_properties(symbol_table=symbol_table)
    )

    for our_type in symbol_table.our_types:
        if (
            isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            )
            and our_type.is_implementation_specific
        ):
            implementation_key = specific_implementations.ImplementationKey(
                f"{our_type.name}.json"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                errors.append(
                    Error(
                        our_type.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific class: {implementation_key}",
                    )
                )
                continue

            try:
                # noinspection PyTypeChecker
                extension = json.loads(code, object_pairs_hook=collections.OrderedDict)
            except Exception as err:
                errors.append(
                    Error(
                        our_type.parsed.node,
                        f"Failed to parse the JSON out of "
                        f"the specific implementation {implementation_key}: {err}",
                    )
                )
                continue

            if not isinstance(extension, dict):
                errors.append(
                    Error(
                        our_type.parsed.node,
                        f"Expected the implementation-specific snippet "
                        f"at {implementation_key} to be a JSON object, "
                        f"but got: {type(extension)}",
                    )
                )
                continue

            update_error = definitions.update_for(
                our_type=our_type, extension=extension
            )
            if update_error is not None:
                errors.append(update_error)

        else:
            if isinstance(our_type, intermediate.Enumeration):
                update_error = definitions.update_for(
                    our_type=our_type,
                    extension=_define_for_enumeration(enumeration=our_type),
                )
                if update_error is not None:
                    errors.append(update_error)

            elif isinstance(our_type, intermediate.ConstrainedPrimitive):
                # NOTE (mristin, 2022-02-11):
                # We in-line the constraints from the constrained primitives directly
                # in the properties. We do not want to introduce separate definitions
                # for them as that would make it more difficult for downstream code
                # generators to generate meaningful code (*e.g.*, code generators for
                # OpenAPI3).
                continue

            elif isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                if len(our_type.concrete_descendants) > 0:
                    # region Inheritable
                    inheritable, definition_errors = _generate_inheritable_definition(
                        cls=our_type,
                        constraints_by_property=constraints_by_class[our_type],
                        fix_pattern=fix_pattern,
                    )

                    if definition_errors is not None:
                        errors.extend(definition_errors)
                        continue

                    assert inheritable is not None
                    update_error = definitions.update_for(
                        our_type=our_type, extension=inheritable
                    )
                    if update_error is not None:
                        errors.append(update_error)
                    # endregion

                    # region Choice
                    if isinstance(our_type, intermediate.ConcreteClass) or (
                        isinstance(our_type, intermediate.AbstractClass)
                        and id(our_type) in ids_of_our_types_in_properties
                    ):
                        update_error = definitions.update_for(
                            our_type=our_type,
                            extension=_generate_choice_definition(cls=our_type),
                        )
                        if update_error is not None:
                            errors.append(update_error)
                    # endregion

                if isinstance(our_type, intermediate.ConcreteClass):
                    definition, definition_errors = _generate_concrete_definition(
                        cls=our_type,
                        constraints_by_property=constraints_by_class[our_type],
                        fix_pattern=fix_pattern,
                    )
                    if definition_errors is not None:
                        errors.extend(definition_errors)
                        continue

                    assert definition is not None

                    update_error = definitions.update_for(
                        our_type=our_type, extension=definition
                    )
                    if update_error is not None:
                        errors.append(update_error)
                else:
                    assert isinstance(our_type, intermediate.AbstractClass)

                    # We do not generate any concrete definition for an abstract class.
                    pass
            else:
                assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    model_types = sorted(
        naming.json_model_type(cls.name)
        for cls in symbol_table.concrete_classes
        if cls.serialization.with_model_type
    )  # type: List[Identifier]

    definitions.update(
        {
            "ModelType": collections.OrderedDict(
                [("type", "string"), ("enum", model_types)]
            )
        }
    )

    definitions_mapping = definitions.get()

    schema["definitions"] = collections.OrderedDict(
        [
            (name, definitions_mapping[name])
            for name in sorted(definitions_mapping.keys())
        ]
    )

    return Stripped(json.dumps(schema, indent=2)), None


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    code, errors = generate(
        symbol_table=context.symbol_table,
        spec_impls=context.spec_impls,
        fix_pattern=fix_pattern_for_utf16,
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the JSON Schema "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "schema.json"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the JSON schema to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



"""Generate implementations and schemas based on an AAS meta-model."""

__version__ = "0.0.19"
__author__ = "Marko Ristin, Nico Braunisch"
__license__ = "License :: OSI Approved :: MIT License"
__status__ = "Production/Stable"
